﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Media;
using System.Threading;
using System.Windows;
using System.Windows.Media.Imaging;
using System.Windows.Threading;
using DnetForm = System.Windows.Forms;

namespace CheckHobbyProcess
{
	/// <summary>
	/// App.xaml の相互作用ロジック
	/// </summary>
	public partial class App : Application
	{
		public static string CAUTION_IMAGE_PATH = "caution.png";
		public static string CAUTION_SOUND_PATH = "caution.wav";
		private static Mutex _mutex = new Mutex(false, "Mutex_CheckHobbyProcess_Opweh0AhGYlg7vweh0AhGYlx8zweh0AhGYlxaEweh0AhGYlaXIweh0AhGYlJ");

		private DispatcherTimer _timer;
		private DnetForm.NotifyIcon _notifyIcon;
		private ConfigWindow _configWindow;
		private List<Process> _hobbyProcessList;

		public BitmapSource CautionBitmap { get; set; }
		public SoundPlayer CautionSound { get; set; }

		private string _cautionImagePath;
		public string CautionImagePath
		{
			get
			{
				return _cautionImagePath;
			}
			set
			{
				try
				{
					Debug.Assert(MainWindow != null);

					FileInfo fileInfo = new FileInfo(Environment.GetCommandLineArgs()[0]);
					string path = fileInfo.Directory.FullName + '\\' + value;
					using(Stream stream = new FileStream(
						path,
						FileMode.Open,
						FileAccess.Read,
						FileShare.ReadWrite | FileShare.Delete
					))
					{
						// ロックしないように指定したstreamを使用する。
						BitmapDecoder decoder = BitmapDecoder.Create(
							stream,
							BitmapCreateOptions.None,
							BitmapCacheOption.Default
						);
						// WriteableBitmapだとすぐに読み込める。
						// BitmapFrame.Create(stream)だと非同期読み込みなのでusing中に読み込めない。
						CautionBitmap = new WriteableBitmap(decoder.Frames[0]);
						CautionBitmap.Freeze();
					}
					// このやり方だとファイルをロックしてしまう
					//CautionBitmap = BitmapFrame.Create(new Uri(value, UriKind.RelativeOrAbsolute));

					((MainWindow)MainWindow).mainImage.Source = CautionBitmap;
					_cautionImagePath = value;
				}
				catch
				{
					_cautionImagePath = null;
				}
			}
		}

		private string _cautionSoundPath;
		public string CautionSoundPath
		{
			get
			{
				return _cautionSoundPath;
			}
			set
			{
				try
				{
					FileInfo appFileInfo = new FileInfo(Environment.GetCommandLineArgs()[0]);
					string wavPath = appFileInfo.Directory.FullName + '\\' + value;
					CautionSound = new SoundPlayer(wavPath);
					_cautionSoundPath = value;
				}
				catch
				{
					_cautionSoundPath = null;
				}
			}
		}

		public void RestartTimer()
		{
			RestartTimer(AppConfig.Current.Interval, false);
		}

		public void Extend()
		{
			RestartTimer(AppConfig.Current.ExtensionOfTime, true);
		}

		public void StopHobbyProcesses()
		{
			List<Process> hobbyProcessList = _hobbyProcessList;

			foreach (Process hobbyProcess in hobbyProcessList)
			{
				try
				{
					hobbyProcess.CloseMainWindow();
					hobbyProcess.Close();
				}
				catch(Exception exc)
				{
					Debug.WriteLine("[" + exc.Message + "]\n" + exc.StackTrace);
				}
			}
		}

		private void OnStartup(object sender, StartupEventArgs e)
		{
			if (!_mutex.WaitOne(0, false))
			{
				Shutdown();
				return;
			}

			string err = null;
			LoadConfig(out err);

			_hobbyProcessList = new List<Process>();

			MainWindow = new MainWindow();
			MainWindow.Show();

			_timer = new DispatcherTimer();
			_timer.Interval = TimeSpan.FromSeconds(AppConfig.Current.Interval);
			_timer.Tick += OnTimerTick;
			_timer.Start();

			CautionImagePath = CAUTION_IMAGE_PATH;
			CautionSoundPath = CAUTION_SOUND_PATH;

			InitializeButtonText();
			InitializeTaskTray();

			if (e.Args.Length == 1 && e.Args[0] == "test")
				MainWindow.Visibility = Visibility.Visible;

			if (err != null)
				MessageBox.Show(MainWindow, err, MainWindow.Title, MessageBoxButton.OK, MessageBoxImage.Information);
		}

		private void OnExit(object sender, ExitEventArgs e)
		{
			if (_notifyIcon != null)
				_notifyIcon.Dispose();
		}

		private void OnTimerTick(object sender, EventArgs e)
		{
			AppConfig conf = AppConfig.Current;

			bool[][] checkPeriodOfTime = conf.CheckPeriodOfTime;
			if (IsCheckPeriodOfTime(checkPeriodOfTime))
			{
				List<string> targetProcessNames = conf.TargetProcessNames;
				List<string> targetPathNames = conf.TargetPathNames;
				string[] untouchableProcessNames = conf.UntouchableProcessNames;
				_hobbyProcessList.Clear();
				Process[] processes = Process.GetProcesses();

				bool found = false;

				foreach (Process p in processes)
				{
					if (IsTarget(p, targetProcessNames, targetPathNames, untouchableProcessNames))
					{
						_hobbyProcessList.Add(p);
						found = true;
					}
				}

				if (found && MainWindow.Visibility != Visibility.Visible)
				{
					MainWindow.Visibility = Visibility.Visible;
					try
					{
						if (CautionSound != null)
						{
							CautionSound.Play();
						}
					}
					catch{}
				}
			}

			// 猶予ボタンで一時的にタイマーの間隔がExtensionOfTimeに書き換えられる。
			// それを戻すためタイマーの間隔をIntervalに書き直す。
			// 場合分けが面倒なので毎回書き換え。
			RestartTimer();
		}

		private void OnExitMenuClick(object sender, EventArgs e)
		{
			Shutdown();
		}

		private void OnConfigMenuClick(object sender, EventArgs e)
		{
			if (_configWindow != null)
			{
				_configWindow.Activate();
				return;
			}

			_timer.Stop();

			if (MainWindow != null)
				MainWindow.Visibility = Visibility.Hidden;

			_configWindow = new ConfigWindow();
			if (_configWindow.ShowDialog() == true)
			{
				try
				{
					AppConfig.Save();
				}
				catch
				{
					MessageBox.Show("設定ファイルの保存に失敗しました。", "「" + MainWindow.Title + "」の設定", MessageBoxButton.OK, MessageBoxImage.Information);
				}
			}
			_configWindow = null;
			RestartTimer();
		}

		private void InitializeButtonText()
		{
			AppConfig conf = AppConfig.Current;
			MainWindow mainWindow = (MainWindow)MainWindow;

			mainWindow.btnStop.Content = conf.StopButtonText;
			mainWindow.btnStop.ToolTip = conf.StopButtonTooltip;
			mainWindow.btnExtend.Content = conf.ExtendButtonText;
			mainWindow.btnExtend.ToolTip = conf.ExtendButtonTooltip;
		}

		private void InitializeTaskTray()
		{
			_notifyIcon = new DnetForm.NotifyIcon();
			_notifyIcon.Text = MainWindow.Title;
			_notifyIcon.Icon = CheckHobbyProcess.Properties.Resources.ICON_TASKTRAY;
			_notifyIcon.Visible = true;

			DnetForm.ContextMenuStrip contextMenu = new DnetForm.ContextMenuStrip();

			DnetForm.ToolStripMenuItem menuItemConfig = new DnetForm.ToolStripMenuItem();
			menuItemConfig.Text = "設定";
			menuItemConfig.Click += OnConfigMenuClick;
			contextMenu.Items.Add(menuItemConfig);

			DnetForm.ToolStripMenuItem menuItemExit = new DnetForm.ToolStripMenuItem();
			menuItemExit.Text = "終了";
			menuItemExit.Click += OnExitMenuClick;
			contextMenu.Items.Add(menuItemExit);

			_notifyIcon.ContextMenuStrip = contextMenu;
		}

		private bool IsCheckPeriodOfTime(bool[][] checkPeriodOfTime)
		{
			DateTime time = DateTime.Now;

			int dayOfWeek = (int)time.DayOfWeek;
			int hour = time.Hour;

			return checkPeriodOfTime[dayOfWeek][hour];
		}

		private bool IsTarget(Process p, List<string> targetProcessNames, List<string> targetPathNames, string[] untouchableProcessNames)
		{
			if (untouchableProcessNames.Contains<string>(p.ProcessName))
				return false;

			if (targetProcessNames.Contains<string>(p.ProcessName))
				return true;

			try
			{
				string modulePath = p.MainModule.FileName;
				foreach (string targetPath in targetPathNames)
				{
					if (modulePath.StartsWith(targetPath))
					{
						return true;
					}
				}
			}
			catch (Exception exc)
			{
				Debug.WriteLine(exc.Message + " : " + p.ProcessName);
			}
			return false;
		}

		private void LoadConfig(out string err)
		{
			err = null;

			try
			{
				AppConfig.Load();
			}
			catch
			{
				try
				{
					AppConfig.Save();
					err = "設定ファイルが読み込めませんでした。新しい設定ファイルを作成しました。設定ファイルのパスは次のとおりです。\n\n"
							+ AppConfig.FilePath;
				}
				catch
				{
					err = "設定ファイルが読み込めませんでした。新しい設定ファイルの作成にも失敗しました。";
				}
			}
		}

		private void RestartTimer(int interval, bool timeIsInMinutes)
		{
			Debug.Assert(_timer != null);

			TimeSpan t = timeIsInMinutes ?
				TimeSpan.FromMinutes(interval) :
				TimeSpan.FromSeconds(interval);


			_timer.Stop();
			_timer.Interval = t;
			_timer.Start();
		}
	}
}
